"""
import_installed - Utility function for testing the installed version of the
package.
"""

from __future__ import absolute_import, print_function

import sys
import os

__all__ = ['import_installed']


def import_installed(module_name):
    """
    Import a Python module/package, controlling whether it is loaded from the
    normal Python module search path, or from an installed version (excluding
    the module in the current directory).

    The TEST_INSTALLED environment variable controls this as follows:

      * If not set or empty, the normal Python module search path is used.
        Because that search path contains the current directory in front of the
        list, this will cause a module directory in the current directory to
        have precedence over any installed versions of the module.

      * If non-empty, the current directory is removed from the Python module
        search path, and an installed version of the module is thus used, even
        when a module directory exists in the current directory. This can be
        used for testing an OS-installed version of the module.

    Example usage, e.g. in a test program::

        from ...utils import import_installed
        xyz = import_installed('xyz')  # pylint: disable=invalid-name
        from xyz import ...

    The number of dots in `from ..utils` depends on where the test program
    containing this code is located, relative to the tests/utils.py file.
    """
    test_installed = os.getenv('TEST_INSTALLED')
    if test_installed:

        # Remove '' directory.
        dirpath = ''
        try:
            ix = sys.path.index(dirpath)
        except ValueError:
            ix = None
        if ix is not None:
            if test_installed == 'DEBUG':
                print("Debug: Removing {0} at index {1} from module search "
                      "path".format(dirpath, ix))
            del sys.path[ix]

        # Move CWD to end. Reason is that when testing with an editable
        # installation, the CWD is needed, but when testing with a non-editable
        # installation, the package should not be found inthe CWD.
        # Note that somehow the CWD gets inserted at the begin of the search
        # path every time, so we need a loop.
        dirpath = os.getcwd()
        while True:
            try:
                ix = sys.path.index(dirpath)
            except ValueError:
                if test_installed == 'DEBUG':
                    print("Debug: Appending {0} to end of module search "
                          "path".format(dirpath))
                sys.path.append(dirpath)
                break
            if ix == len(sys.path) - 1:
                # it exists once at the end
                break
            if test_installed == 'DEBUG':
                print("Debug: Removing {0} at index {1} from module search "
                      "path".format(dirpath, ix))
            del sys.path[ix]

    if module_name not in sys.modules:
        module = __import__(module_name, level=0)  # only absolute imports
        if test_installed == 'DEBUG':
            print("Debug: {0} module newly loaded from: {1}".
                  format(module_name, module.__file__))
    else:
        module = sys.modules[module_name]
        if test_installed == 'DEBUG':
            print("Debug: {0} module was already loaded from: {1}".
                  format(module_name, module.__file__))
    return module



"""
This module provides class NocaseList.
"""

import sys
import os
from typing import Callable, AnyStr, Optional, Union
try:
    from typing import SupportsIndex  # type: ignore
except ImportError:
    from typing_extensions import SupportsIndex  # Python <=3.7
try:
    from typing import TypeAlias  # type: ignore
except ImportError:
    from typing_extensions import TypeAlias  # Python <=3.9
if sys.version_info[0:2] >= (3, 9):
    from collections.abc import Iterable  # type: ignore
else:
    # Before py39, collections.abc.Iterable did not support generic type
    from typing import Iterable

__all__ = ['NocaseList']

# This env var is set when building the docs. It causes the methods
# that are supposed to exist only in a particular Python version, not to be
# removed, so they appear in the docs.
BUILDING_DOCS = os.environ.get('BUILDING_DOCS', False)

# Type for values in NocaseList
Value: TypeAlias = Optional[AnyStr]

# Type for returning single values or slices
ValueOrNocaseList: TypeAlias = Union[Value, 'NocaseList']

# Type for other argument in rich comparison methods
OtherList: TypeAlias = Iterable[Value]  # pylint: disable=unsubscriptable-object

# Type for index argument that can also specify a slice
IndexOrSlice: TypeAlias = Union[SupportsIndex, slice]


class NocaseList(list):
    """
    A case-insensitive and case-preserving list.

    The list is case-insensitive: Whenever items of the list are looked up by
    value or item values are compared, that is done case-insensitively. The
    case-insensitivity is defined by performing the lookup or comparison on the
    result of the :meth:`__casefold__` method on the on list items.
    `None` is allowed as a list value and will not be case folded.

    The list is case-preserving: Whenever the value of list items is returned,
    they have the lexical case that was originally specified when adding
    or updating the item.

    Except for the case-insensitivity of its items, it behaves like, and is
    in fact derived from, the built-in :class:`py:list` class in order to
    facilitate type checks.

    The implementation maintains a second list with the casefolded items of
    the inherited list, and ensures that both lists are in sync.

    The list supports serialization via the Python :mod:`py:pickle` module.
    To save space and time, only the originally cased list is serialized.
    """

    # Methods not implemented:
    #
    # * __getattribute__(self, name): The method inherited from object is used;
    #   no reason to have a different implementation.
    #
    # * __sizeof__(self): The method inherited from list is used; no reason
    #   to have a different implementation.
    #
    # * __len__(self): The method inherited from list is used; no reason
    #   to have a different implementation.
    #
    # __repr__(): The method inherited from list is used; no reason
    #   to have a different implementation.
    #
    # __getitem__(): The method inherited from list is used; no reason
    #   to have a different implementation.
    #
    # __iter__(): The method inherited from list is used; no reason
    #   to have a different implementation.

    def __init__(self, iterable=()) -> None:
        """
        Initialize the list with the items in the specified iterable.
        """
        super(NocaseList, self).__init__(iterable)

        # The _casefolded_list attribute is a list with the same items as the
        # original (inherited) list, except they are casefolded using the
        # __casefold__() method.

        # The following is an optimization based on the assumption that in
        # many cases, casefolding the input list is more expensive than
        # copying it (plus the overhead to check that).
        if isinstance(iterable, NocaseList):
            # pylint: disable=protected-access
            casefolded_list = iterable._casefolded_list.copy()
        else:
            casefolded_list = self._new_casefolded_list(self)
        self._casefolded_list: list = casefolded_list

    def _new_casefolded_list(self, lst: OtherList) -> list:
        """
        Return a casefolded list from the input list.
        """
        result = []
        for value in lst:
            result.append(self._casefolded_value(value))
        return result

    def _casefolded_value(self, value: Value) -> Value:
        """
        This method returns the casefolded value and handles the case of value
        being `None`. The value may be a string or an list/tuple of strings.
        """
        if value is None:
            return None
        if isinstance(value, (list, tuple)):
            return [self._casefolded_value(v) for v in value]
        return self.__casefold__(value)

    @staticmethod
    def __casefold__(value: AnyStr) -> AnyStr:
        """
        This method implements the case-insensitive behavior of the class.

        It returns a case-insensitive form of the input value by calling a
        "casefold method" on the value. The input value will not be `None`.

        The casefold method called by this method is :meth:`py:str.casefold`.
        If that method does not exist on the key value (e.g. because it is a
        byte string), :meth:`py:bytes.lower` is called, for compatibility with
        earlier versions of the package.

        This method can be overridden by users in order to change the
        case-insensitive behavior of the class.
        See :ref:`Overriding the default casefold method` for details.

        Parameters:
          value (str or bytes): Input value. Will not be `None`.

        Returns:
          str or bytes: Case-insensitive form of the input value.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        try:
            return value.casefold()  # type: ignore
        except AttributeError:
            return value.lower()

    def __getstate__(self):
        """
        Called when pickling the object, see :meth:`py:object.__getstate__`.

        In order to save space and time, only the list with the originally
        cased items is saved, but not the second list with the casefolded
        items.
        """
        # This copies the state of the inherited list even though it is
        # not visible in self.__dict__.
        state = self.__dict__.copy()
        del state['_casefolded_list']
        return state

    def __setstate__(self, state):
        """
        Called when unpickling the object, see :meth:`py:object.__setstate__`.
        """
        self.__dict__.update(state)
        self._casefolded_list = self._new_casefolded_list(self)

    def __setitem__(self, index: IndexOrSlice, value: Value) -> None:
        """
        Update the value of the item at an existing index or slice in the list.

        Invoked using ``ncl[index] = value``.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        super(NocaseList, self).__setitem__(index, value)  # type: ignore
        self._casefolded_list[index] = self._casefolded_value(value)  # type: ignore # noqa: E501 pylint: disable=line-too-long

    def __delitem__(self, index: IndexOrSlice) -> None:
        """
        Delete an item at an existing index or slice from the list.

        Invoked using ``del ncl[index]``.
        """
        super(NocaseList, self).__delitem__(index)
        del self._casefolded_list[index]

    def __contains__(self, value: Value) -> bool:
        """
        Return a boolean indicating whether the list contains at least one
        item with the value, by looking it up case-insensitively.

        Invoked using ``value in ncl``.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        return self._casefolded_value(value) in self._casefolded_list

    def __add__(self, other: OtherList) -> 'NocaseList':
        """
        Return a new :class:`NocaseList` object that contains the items from
        the left hand operand (``self``) and the items from the right hand
        operand (``other``).

        The right hand operand (``other``) must be an instance of
        :class:`py:list` (including :class:`NocaseList`) or :class:`py:tuple`.
        The operands are not changed.

        Invoked using e.g. ``ncl + other``

        Raises:
          TypeError: The other iterable is not a list or tuple
        """
        if not isinstance(other, (list, tuple)):
            raise TypeError(
                "Can only concatenate list or tuple (not {t}) to NocaseList".
                format(t=type(other)))
        lst = self.copy()
        lst.extend(other)
        return lst

    def __iadd__(self, other: OtherList) -> 'NocaseList':
        """
        Extend the left hand operand (``self``) by the items from the right
        hand operand (``other``).

        The ``other`` parameter must be an iterable but is otherwise not
        restricted in type. Thus, if it is a string, the characters of the
        string are added as distinct items to the list.

        Invoked using ``ncl += other``.
        """
        # Note: It is unusual that the method has to return self, but it was
        # verified that this is necessary.
        self.extend(other)
        return self

    def __mul__(self, number: int) -> 'NocaseList':  # type: ignore
        """
        Return a new :class:`NocaseList` object that contains the items from
        the left hand operand (``self``) as many times as specified by the right
        hand operand (``number``).

        A number <= 0 causes the returned list to be empty.

        The left hand operand (``self``) is not changed.

        Invoked using ``ncl * number``.
        """
        # Despite using type hints, passing a non-int object does not raise
        # any exception, so we still need to check the type:
        if not isinstance(number, int):
            raise TypeError(
                "Cannot multiply NocaseList by non-integer of type {t}".
                format(t=type(number)))
        lst = NocaseList()
        for _ in range(0, number):
            lst.extend(self)
        return lst

    def __rmul__(self, number: int) -> 'NocaseList':  # type: ignore
        """
        Return a new :class:`NocaseList` object that contains the items from
        the right hand operand (``self``) as many times as specified by the left
        hand operand (``number``).

        A number <= 0 causes the returned list to be empty.

        The right hand operand (``self``) is not changed.

        Invoked using ``number * ncl``.
        """
        lst = self * number  # Delegates to __mul__()
        return lst

    def __imul__(self, number: int) -> 'NocaseList':  # type: ignore
        """
        Change the left hand operand (``self``) so that it contains the items
        from the original left hand operand (``self``) as many times as
        specified by the right hand operand (``number``).

        A number <= 0 will empty the left hand operand.

        Invoked using ``ncl *= number``.
        """
        # Despite using type hints, passing a non-int object does not raise
        # any exception, so we still need to check the type:
        if not isinstance(number, int):
            raise TypeError(
                "Cannot multiply NocaseList by non-integer of type {t}".
                format(t=type(number)))
        if number <= 0:
            del self[:]
            del self._casefolded_list[:]
        else:
            self_items = list(self)
            for _ in range(0, number - 1):
                self.extend(self_items)
        # Note: It is unusual that the method has to return self, but it was
        # verified that this is necessary.
        return self

    def __reversed__(self) -> 'NocaseList':  # type: ignore
        """
        Return a shallow copy of the list that has its items reversed in order.

        Invoked using ``reversed(ncl)``.
        """
        lst = self.copy()
        lst.reverse()
        return lst

    def __eq__(self, other: object) -> bool:
        """
        Return a boolean indicating whether the list and the other list are
        equal, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl == other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            # pylint: disable=protected-access
            return self._casefolded_list == other._casefolded_list

        if isinstance(other, Iterable):
            return self._casefolded_list == self._new_casefolded_list(other)

        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """
        Return a boolean indicating whether the list and the other list are
        not equal, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl != other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return NotImplemented
        return not eq

    def __gt__(self, other: object) -> bool:
        """
        Return a boolean indicating whether the list is greater than the other
        list, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl > other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            # pylint: disable=protected-access
            return self._casefolded_list > other._casefolded_list

        if isinstance(other, Iterable):
            return self._casefolded_list > self._new_casefolded_list(other)

        return NotImplemented

    def __lt__(self, other: object) -> bool:
        """
        Return a boolean indicating whether the list is less than the other
        list, by comparing corresponding list items case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl < other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        if isinstance(other, NocaseList):
            # pylint: disable=protected-access
            return self._casefolded_list < other._casefolded_list

        if isinstance(other, Iterable):
            return self._casefolded_list < self._new_casefolded_list(other)

        return NotImplemented

    def __ge__(self, other: object) -> bool:
        """
        Return a boolean indicating whether the list is greater than or
        equal to the other list, by comparing corresponding list items
        case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl >= other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        lt = self.__lt__(other)
        if lt is NotImplemented:
            return NotImplemented
        return not lt

    def __le__(self, other: object) -> bool:
        """
        Return a boolean indicating whether the list is less than or
        equal to the other list, by comparing corresponding list items
        case-insensitively.

        The other list may be a :class:`NocaseList` object or any other
        iterable. In all cases, the comparison takes place case-insensitively.

        Invoked using e.g. ``ncl <= other``.

        Raises:
          AttributeError: A value in the other list does not have the casefold
            method.
        """
        gt = self.__gt__(other)
        if gt is NotImplemented:
            return NotImplemented
        return not gt

    def count(self, value: Value) -> int:
        """
        Return the number of times the specified value occurs in the list,
        comparing the value and the list items case-insensitively.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        return self._casefolded_list.count(self._casefolded_value(value))

    def copy(self) -> 'NocaseList':
        """
        Return a shallow copy of the list.
        """
        return NocaseList(self)

    def clear(self) -> None:
        """
        Remove all items from the list (and return None).
        """
        super(NocaseList, self).clear()
        self._casefolded_list.clear()

    def index(self, value: Value, start: SupportsIndex = 0,
              stop: SupportsIndex = 9223372036854775807) -> int:
        """
        Return the index of the first item that is equal to the specified
        value, comparing the value and the list items case-insensitively.

        The search is limited to the index range defined by the specified
        ``start`` and ``stop`` parameters, whereby ``stop`` is the index
        of the first item after the search range.

        Raises:
          AttributeError: The value does not have the casefold method.
          ValueError: No such item is found.
        """
        return self._casefolded_list.index(
            self._casefolded_value(value), start, stop)

    def append(self, value: Value) -> None:
        """
        Append the specified value as a new item to the end of the list
        (and return None).

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        super(NocaseList, self).append(value)
        self._casefolded_list.append(self._casefolded_value(value))

    def extend(self, values: Iterable) -> None:
        """
        Extend the list by the items in the specified iterable
        (and return None).

        Raises:
          AttributeError: A value in the iterable does not have the casefold
            method.
        """
        super(NocaseList, self).extend(values)
        # The following is a circumvention for a behavior of the 'pickle' module
        # that during unpickling may call this method on an object that has
        # been created with __new__() without calling __init__().
        try:
            for value in values:
                self._casefolded_list.append(self._casefolded_value(value))
        except AttributeError:
            self._casefolded_list = self._new_casefolded_list(self)

    def insert(self, index: SupportsIndex, value: Value) -> None:
        """
        Insert a new item with specified value before the item at the specified
        index (and return None).

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        super(NocaseList, self).insert(index, value)
        self._casefolded_list.insert(index, self._casefolded_value(value))

    def pop(self, index: SupportsIndex = -1) -> Value:
        """
        Return the value of the item at the specified index and also remove it
        from the list.
        """
        self._casefolded_list.pop(index)
        return super(NocaseList, self).pop(index)

    def remove(self, value: Value) -> None:
        """
        Remove the first item from the list whose value is equal to the
        specified value (and return None), comparing the value and the list
        items case-insensitively.

        Raises:
          AttributeError: The value does not have the casefold method.
        """
        self._casefolded_list.remove(self._casefolded_value(value))
        super(NocaseList, self).remove(value)

    def reverse(self) -> None:
        """
        Reverse the items in the list in place (and return None).
        """
        super(NocaseList, self).reverse()
        self._casefolded_list = self._new_casefolded_list(self)

    def sort(self, *, key: Optional[Callable] = None,
             reverse: bool = False) -> None:
        """
        Sort the items in the list in place (and return None).

        The sort is stable, in that the order of two (case-insensitively) equal
        elements is maintained.

        By default, the list is sorted in ascending order of its casefolded
        item values. If a key function is given, it is applied once to each
        casefolded list item and the list is sorted in ascending or
        descending order of their key function values.

        The ``reverse`` flag can be set to sort in descending order.
        """

        def casefolded_key(value):
            """Key function used for sorting"""
            # The function cannot raise AttributeError due to missing casefold
            # method, because the list items have been verified for that when
            # adding them to the list.
            if key:
                return key(self._casefolded_value(value))
            return self._casefolded_value(value)

        super(NocaseList, self).sort(key=casefolded_key, reverse=reverse)
        self._casefolded_list = self._new_casefolded_list(self)



"""
Test the NocaseList class.
"""

from __future__ import absolute_import

import sys
import os
import re
import unicodedata
import pickle
import pytest
import six  # type: ignore

from ..utils.simplified_test_function import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ..utils.import_installed import import_installed
nocaselist = import_installed('nocaselist')
from nocaselist import NocaseList as _NocaseList  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# pylint: disable=use-dict-literal

# Controls whether the tests are run against a standard dict instead.
TEST_AGAINST_LIST = bool(os.getenv('TEST_AGAINST_LIST'))

if TEST_AGAINST_LIST:
    print("\nInfo: test_nocaselist.py tests run against standard list")

# The list class being tested
# pylint: disable=invalid-name
NocaseList = list if TEST_AGAINST_LIST else _NocaseList

# Flag indicating that standard dict is guaranteed to preserve order
DICT_PRESERVES_ORDER = sys.version_info[0:2] >= (3, 7)


def ensure_unicode(value):
    """
    Return the value as a unicode string.
    """
    if value is None:
        return None
    if isinstance(value, six.text_type):
        return value
    return value.decode('utf-8')


def assert_equal(list1, list2, verify_order=True):
    """
    Assert that list1 is equal to list2.
    Check consistency of list1.

    list1: Must be a NocaseList or list object.
    list2: Must be a NocaseList or list or iterable object.
    """

    list1_lst = list(list1)  # Uses NocaseList.__iter__()
    list2_lst = list(list2)  # Uses NocaseList.__iter__()

    if verify_order:
        assert list1_lst == list2_lst
    else:
        assert sorted(list1_lst, key=ensure_unicode) == \
            sorted(list2_lst, key=ensure_unicode)

    # Check consistency of list1 between its internal casefolded list and
    # its external list
    if isinstance(list1, _NocaseList):
        list1_cf = list1._casefolded_list  # pylint: disable=protected-access
        assert len(list1_lst) == len(list1_cf)
        for i, value in enumerate(list1):  # Uses NocaseList.__iter__()
            value_cf = list1_cf[i]
            assert list1.__casefold__(value) == value_cf


TESTCASES_NOCASELIST_INIT = [

    # Testcases for NocaseList.__init__() / ncl=NocaseList()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to NocaseList().
    #   * init_kwargs: Dict of keyword arguments to NocaseList().
    #   * exp_list: Expected resulting list.
    #   * verify_order: Flag to verify the expected order.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list from no args",
        dict(
            init_args=(),
            init_kwargs={},
            exp_list=[],
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty list from empty list as positional arg",
        dict(
            init_args=([],),
            init_kwargs={},
            exp_list=[],
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty list from empty tuple as positional arg",
        dict(
            init_args=(tuple(),),
            init_kwargs={},
            exp_list=[],
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty list from empty dict as positional arg",
        dict(
            init_args=({},),
            init_kwargs={},
            exp_list=[],
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty list from empty NocaseList as positional arg",
        dict(
            init_args=(NocaseList(),),
            init_kwargs={},
            exp_list=[],
            verify_order=True,
        ),
        None, None, True
    ),

    # Non-empty NocaseList
    (
        "List from list as positional arg",
        dict(
            init_args=(['Dog', b'Cat'],),
            init_kwargs={},
            exp_list=['Dog', b'Cat'],
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "List from tuple as positional arg",
        dict(
            init_args=(('Dog', b'Cat'),),
            init_kwargs={},
            exp_list=['Dog', b'Cat'],
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "List from dict as positional arg (uses only the keys)",
        dict(
            init_args=({'Dog': 'Kitten', b'Cat': 'Budgie'},),
            init_kwargs={},
            exp_list=['Dog', b'Cat'],
            verify_order=DICT_PRESERVES_ORDER,
        ),
        None, None, True
    ),
    (
        "List from string as positional arg (string chars become list items)",
        dict(
            init_args=('Dog',),
            init_kwargs={},
            exp_list=['D', 'o', 'g'],
            verify_order=True,
        ),
        None, None, True
    ),

    # Error cases
    (
        "'iterable' keyword arg (no kwargs)",
        dict(
            init_args=(),
            init_kwargs=dict(iterable=[1, 2]),
            exp_list=None,
            verify_order=None,
        ),
        TypeError if TEST_AGAINST_LIST else AttributeError, None, True
        # TODO(issue #25): Make behavior consistent to standard list
    ),
    (
        "None as positional arg (is not iterable)",
        dict(
            init_args=(None,),
            init_kwargs={},
            exp_list=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "Integer type as positional arg",
        dict(
            init_args=(42,),
            init_kwargs={},
            exp_list=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "Two positional args",
        dict(
            init_args=([], []),
            init_kwargs={},
            exp_list=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_INIT)
@simplified_test_function
def test_NocaseList_init(testcase, init_args, init_kwargs, exp_list,
                         verify_order):
    """
    Test function for NocaseList.__init__() / ncl=NocaseList()
    """

    # The code to be tested
    act_list = NocaseList(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify that NocaseList inherits from list
    assert isinstance(act_list, list)

    assert_equal(act_list, exp_list, verify_order)  # Uses NocaseList.__iter__()


TESTCASES_NOCASELIST_REPR = [

    # Testcases for NocaseList.__repr__() / repr(ncl)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list",
        dict(
            nclist=NocaseList(),
        ),
        None, None, True
    ),
    (
        "List with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_REPR)
@simplified_test_function
def test_NocaseList_repr(testcase, nclist):
    """
    Test function for NocaseList.__repr__() / repr(ncl)
    """

    # The code to be tested
    result = repr(nclist)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert re.match(r'^\[.*\]$', result)

    # Note: This only tests for existence of each item, not for excess items
    # or representing the correct order.
    for item in nclist:
        exp_item_result = repr(item)
        assert exp_item_result in result


TESTCASES_NOCASELIST_GETITEM = [

    # Testcases for NocaseList.__getitem__() / val = ncl[idx]

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * index: Index to be used for the test.
    #   * exp_value: Expected value for the index.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with None as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with unicode string as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index='',
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with byte string as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index=b'',
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with non-existing index 0 (out of range)",
        dict(
            nclist=NocaseList(),
            index=0,
            exp_value=None,
        ),
        IndexError, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with None as index (invalid type)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            index=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "List with two items, with string as index (invalid type)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            index='',
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "List with two items, with non-existing index 2 (out of range)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            index=2,
            exp_value=None,
        ),
        IndexError, None, True
    ),
    (
        "List with two items, with index 0",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            index=0,
            exp_value='Dog',
        ),
        None, None, True
    ),
    (
        "List with two items, with index 1",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            index=1,
            exp_value=b'Cat',
        ),
        None, None, True
    ),
    (
        "List with two items, with index -1",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            index=-1,
            exp_value=b'Cat',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_GETITEM)
@simplified_test_function
def test_NocaseList_getitem(testcase, nclist, index, exp_value):
    """
    Test function for NocaseList.__getitem__() / val = ncl[idx]
    """

    # The code to be tested
    act_value = nclist[index]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_value == exp_value, "Unexpected value at index {}".format(index)


TESTCASES_NOCASELIST_SETITEM = [

    # Testcases for NocaseList.__setitem__() / ncl[idx] = value

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * index: Index to be used for the test.
    #   * value: Value to be used for the test.
    #   * exp_nclist: Expected NocaseList object after test function, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with None as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index=None,
            value=None,
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with unicode string as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index='',
            value=None,
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with byte string as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index=b'',
            value=None,
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with non-existing index 0 (out of range)",
        dict(
            nclist=NocaseList(),
            index=0,
            value='Newbie',
            exp_nclist=None,
        ),
        IndexError, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with None as index (invalid type)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            index=None,
            value='Kitten',
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "List with two items, setting unicode string at index 0",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=0,
            value='Newbie',
            exp_nclist=NocaseList(['Newbie', 'Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, setting byte string at index 1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=1,
            value=b'Newbie',
            exp_nclist=NocaseList(['Dog', b'Newbie']),
        ),
        None, None, True
    ),
    (
        "List with two items, setting string at index -1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=-1,
            value='Newbie',
            exp_nclist=NocaseList(['Dog', 'Newbie']),
        ),
        None, None, True
    ),
    (
        "List with two items, setting string at index 2 (out of range)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=2,
            value='Newbie',
            exp_nclist=None,
        ),
        IndexError, None, True
    ),
    (
        "Updating string list at slice 0:2",
        dict(
            nclist=NocaseList(['A1', 'B2', 'C3', 'D4']),
            index=slice(0, 2),
            value=['XX', 'YY'],
            exp_nclist=NocaseList(['XX', 'YY', 'C3', 'D4']),
        ),
        None, None, True
    ),
    (
        "Updating string list at slice 0:2:2",
        dict(
            nclist=NocaseList(['A1', 'B2', 'C3', 'D4']),
            index=slice(0, 4, 2),
            value=['XX', 'YY'],
            exp_nclist=NocaseList(['XX', 'B2', 'YY', 'D4']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_SETITEM)
@simplified_test_function
def test_NocaseList_setitem(testcase, nclist, index, value, exp_nclist):
    """
    Test function for NocaseList.__setitem__() / ncl[idx]=value
    """

    # The code to be tested
    nclist[index] = value

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseList features, but that is
    # unavoidable if we want to work through the public interface:

    act_value = nclist[index]  # Uses NocaseList.__getitem__()

    assert act_value == value, "Unexpected value at index {}".format(index)

    assert_equal(nclist, exp_nclist)  # Uses NocaseList.__iter__()


TESTCASES_NOCASELIST_DELITEM = [

    # Testcases for NocaseList.__delitem__() / del ncl[idx]

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * index: Index to be used for the test.
    #   * exp_nclist: Expected NocaseList object after deletion, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with None as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index=None,
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with string as index (invalid type)",
        dict(
            nclist=NocaseList(),
            index='',
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty list, with non-existing index 0 (out of range)",
        dict(
            nclist=NocaseList(),
            index=0,
            exp_nclist=None,
        ),
        IndexError, None, True
    ),
    (
        "Empty list, with non-existing index -1 (out of range)",
        dict(
            nclist=NocaseList(),
            index=-1,
            exp_nclist=None,
        ),
        IndexError, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with None as index (invalid type)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=None,
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "List with two items, with string as index (invalid type)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index='',
            exp_nclist=None,
        ),
        TypeError, None, True
    ),
    (
        "List with two items, with existing index 0",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=0,
            exp_nclist=NocaseList(['Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, with existing index -1 (last from end)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=-1,
            exp_nclist=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "List with two items, with non-existing index 2 (out of range)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            index=2,
            exp_nclist=None,
        ),
        IndexError, None, True
    ),
    (
        "Deleting string list at slice 0:2",
        dict(
            nclist=NocaseList(['A1', 'B2', 'C3', 'D4']),
            index=slice(0, 2),
            exp_nclist=NocaseList(['C3', 'D4']),
        ),
        None, None, True
    ),
    (
        "Deleting string list at slice 0:2:2",
        dict(
            nclist=NocaseList(['A1', 'B2', 'C3', 'D4']),
            index=slice(0, 4, 2),
            exp_nclist=NocaseList(['B2', 'D4']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_DELITEM)
@simplified_test_function
def test_NocaseList_delitem(testcase, nclist, index, exp_nclist):
    """
    Test function for NocaseList.__delitem__() / del ncl[idx]
    """

    # The code to be tested
    del nclist[index]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_equal(nclist, exp_nclist)  # Uses NocaseList.__iter__()


TESTCASES_NOCASELIST_ITER = [

    # Testcases for NocaseList.__iter__() / for item in ncl

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * exp_items: List with expected items in expected order.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list",
        dict(
            nclist=NocaseList(),
            exp_items=[],
        ),
        None, None, True
    ),
    (
        "List with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            exp_items=['Dog', 'Cat'],
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_ITER)
@simplified_test_function
def test_NocaseList_iter(testcase, nclist, exp_items):
    """
    Test function for NocaseList.__iter__() / for item in ncl
    """

    # The code to be tested
    act_items = []
    for item in nclist:
        act_items.append(item)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_items == exp_items


TESTCASES_NOCASELIST_CONTAINS = [

    # Testcases for NocaseList.__contains__() / value in ncl

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * value: Value to be used for the test.
    #   * exp_result: Expected result of the test function, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with None as value (no casefold method)",
        dict(
            nclist=NocaseList(),
            value=None,
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty list, with integer value 0 (no casefold method)",
        dict(
            nclist=NocaseList(),
            value=0,
            exp_result=False if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Empty list, with non-existing empty unicode string value (not found)",
        dict(
            nclist=NocaseList(),
            value='',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty list, with non-existing empty byte string value (not found)",
        dict(
            nclist=NocaseList(),
            value=b'',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty list, with non-existing non-empty unicode string value "
        "(not found)",
        dict(
            nclist=NocaseList(),
            value='Dog',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty list, with non-existing non-empty byte string value "
        "(not found)",
        dict(
            nclist=NocaseList(),
            value=b'Dog',
            exp_result=False,
        ),
        None, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with None as value (no casefold method)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            value=None,
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "List with two items, with integer value 0 (no casefold method)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            value=0,
            exp_result=False if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "List with two items, with non-existing empty unicode string value "
        "(not found)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            value='',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "List with two items, with non-existing empty byte string value "
        "(not found)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            value=b'',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "List with two items, with non-existing non-empty unicode string value "
        "(not found)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            value='invalid',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "List with two items, with non-existing non-empty byte string value "
        "(not found)",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            value=b'invalid',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing unicode string value in original "
        "case",
        dict(
            nclist=NocaseList(['Dog', b'Cat']),
            value='Dog',
            exp_result=True,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing byte string value in original "
        "case",
        dict(
            nclist=NocaseList([b'Dog', 'Cat']),
            value=b'Dog',
            exp_result=True,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing unicode string value in "
        "non-original upper case",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='DOG',
            exp_result=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing byte string value in "
        "non-original upper case",
        dict(
            nclist=NocaseList([b'Dog', 'Cat']),
            value=b'DOG',
            exp_result=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing unicode string value in "
        "non-original lower case",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='dog',
            exp_result=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing byte string value in "
        "non-original lower case",
        dict(
            nclist=NocaseList([b'Dog', 'Cat']),
            value=b'dog',
            exp_result=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing unicode string value in "
        "non-original mixed case",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='doG',
            exp_result=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with two items, with existing byte string value in "
        "non-original mixed case",
        dict(
            nclist=NocaseList([b'Dog', 'Cat']),
            value=b'doG',
            exp_result=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_CONTAINS)
@simplified_test_function
def test_NocaseList_contains(testcase, nclist, value, exp_result):
    """
    Test function for NocaseList.__contains__() / value in ncl
    """

    # The code to be tested
    act_result = value in nclist

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_result == exp_result, \
        "Unexpected result for value {!r}".format(value)


TESTCASES_NOCASELIST_SIZEOF = [

    # Testcases for NocaseList.__sizeof__() / len(ncl)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * exp_result: Expected result of the test function, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list",
        dict(
            nclist=NocaseList(),
            exp_result=0,
        ),
        None, None, True
    ),
    (
        "List with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            exp_result=2,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_SIZEOF)
@simplified_test_function
def test_NocaseList_sizeof(testcase, nclist, exp_result):
    """
    Test function for NocaseList.__sizeof__() / len(ncl)
    """

    # The code to be tested
    result = len(nclist)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result


TESTCASES_NOCASELIST_ADD = [

    # Testcases for NocaseList.__add__() / ncl + val

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList (LHO) object to be used for the test.
    #   * other: Other iterable (RHO) to be used for the test.
    #   * exp_result: Expected result of the test function, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Add an empty NocaseList and a string (must be list/tuple)",
        dict(
            nclist=NocaseList(),
            other='Dog',
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Add an empty NocaseList and an integer (must be list/tuple)",
        dict(
            nclist=NocaseList(),
            other=42,
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Add an empty NocaseList and a tuple (list: must be list, "
        "NocaseList: success)",
        dict(
            nclist=NocaseList(),
            other=('Dog',),
            exp_result=None if TEST_AGAINST_LIST \
            else NocaseList(['Dog']),
        ),
        TypeError if TEST_AGAINST_LIST else None, None, True
        # This is an extension of NocaseList over list
    ),
    (
        "Add an empty NocaseList and a list with one string item",
        dict(
            nclist=NocaseList(),
            other=['Dog'],
            exp_result=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "Add an empty NocaseList and a NocaseList with one string item",
        dict(
            nclist=NocaseList(),
            other=NocaseList(['Dog']),
            exp_result=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "Add an empty NocaseList and a list with one integer item "
        "(list: success, NocaseList: no casefold)",
        dict(
            nclist=NocaseList(),
            other=[42],
            exp_result=NocaseList([42]) if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Add a NocaseList with two items and a list with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            other=['Budgie', 'Fish'],
            exp_result=NocaseList(['Dog', 'Cat', 'Budgie', 'Fish']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_ADD)
@simplified_test_function
def test_NocaseList_add(testcase, nclist, other, exp_result):
    """
    Test function for NocaseList.__add__() / ncl + val
    """

    org_nclist = NocaseList(nclist)

    # The code to be tested
    result = nclist + other

    # Verify the input NocaseList object has not changed
    assert_equal(nclist, org_nclist)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_equal(result, exp_result)


TESTCASES_NOCASELIST_IADD = [

    # Testcases for NocaseList.__iadd__() / ncl += val

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList (LHO) object to be used for the test.
    #   * other: Other iterable (RHO) to be used for the test.
    #   * exp_result: Expected result of the test function, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Extend an empty NocaseList by a string (string chars become items)",
        dict(
            nclist=NocaseList(),
            other='Dog',
            exp_result=NocaseList(['D', 'o', 'g']),
        ),
        None, None, True
    ),
    (
        "Extend an empty NocaseList by an integer (no iterable)",
        dict(
            nclist=NocaseList(),
            other=42,
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Extend an empty NocaseList by a tuple",
        dict(
            nclist=NocaseList(),
            other=('Dog',),
            exp_result=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "Extend an empty NocaseList by a list with one string item",
        dict(
            nclist=NocaseList(),
            other=['Dog'],
            exp_result=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "Extend an empty NocaseList by a NocaseList with one string item",
        dict(
            nclist=NocaseList(),
            other=NocaseList(['Dog']),
            exp_result=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "Extend an empty NocaseList by a list with one integer item "
        "(list: success, NocaseList: no casefold)",
        dict(
            nclist=NocaseList(),
            other=[42],
            exp_result=NocaseList([42]) if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Extend a NocaseList with two items by a list with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            other=['Budgie', 'Fish'],
            exp_result=NocaseList(['Dog', 'Cat', 'Budgie', 'Fish']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_IADD)
@simplified_test_function
def test_NocaseList_iadd(testcase, nclist, other, exp_result):
    """
    Test function for NocaseList.__iadd__() / ncl += val
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)
    nclist_copy_id = id(nclist_copy)

    # The code to be tested
    nclist_copy += other

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the object has been changed in place
    assert id(nclist_copy) == nclist_copy_id

    assert_equal(nclist_copy, exp_result)


TESTCASES_NOCASELIST_MUL = [

    # Testcases for NocaseList.__mul__() / ncl * num
    # Testcases for NocaseList.__rmul__() / num * ncl
    # Testcases for NocaseList.__imul__() / ncl *= num

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * number: Number parameter for the test function.
    #   * exp_result: Expected result of the test function, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list times -1",
        dict(
            nclist=NocaseList(),
            number=-1,
            exp_result=NocaseList(),
        ),
        None, None, True
    ),
    (
        "Empty list times 0",
        dict(
            nclist=NocaseList(),
            number=0,
            exp_result=NocaseList(),
        ),
        None, None, True
    ),
    (
        "Empty list times 1",
        dict(
            nclist=NocaseList(),
            number=1,
            exp_result=NocaseList(),
        ),
        None, None, True
    ),
    (
        "Empty list times 2",
        dict(
            nclist=NocaseList(),
            number=2,
            exp_result=NocaseList(),
        ),
        None, None, True
    ),
    (
        "Empty list times -0.5 (must be integer)",
        dict(
            nclist=NocaseList(),
            number=-0.5,
            exp_result=NocaseList(),
        ),
        TypeError, None, True
    ),
    (
        "Empty list times +0.5 (must be integer)",
        dict(
            nclist=NocaseList(),
            number=+0.5,
            exp_result=NocaseList(),
        ),
        TypeError, None, True
    ),
    (
        "List with two items times -1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            number=-1,
            exp_result=NocaseList(),
        ),
        None, None, True
    ),
    (
        "List with two items times 0",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            number=0,
            exp_result=NocaseList(),
        ),
        None, None, True
    ),
    (
        "List with two items times 1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            number=1,
            exp_result=NocaseList(['Dog', 'Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items times 2",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            number=2,
            exp_result=NocaseList(['Dog', 'Cat', 'Dog', 'Cat']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_MUL)
@simplified_test_function
def test_NocaseList_mul(testcase, nclist, number, exp_result):
    """
    Test function for NocaseList.__mul__() / ncl * num
    """

    org_nclist = NocaseList(nclist)

    # The code to be tested
    result = nclist * number

    # Verify the input NocaseList object has not changed
    assert_equal(nclist, org_nclist)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_equal(result, exp_result)


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_MUL)
@simplified_test_function
def test_NocaseList_rmul(testcase, nclist, number, exp_result):
    """
    Test function for NocaseList.__rmul__() / num * ncl
    """

    org_nclist = NocaseList(nclist)

    # The code to be tested
    result = number * nclist

    # Verify the input NocaseList object has not changed
    assert_equal(nclist, org_nclist)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_equal(result, exp_result)


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_MUL)
@simplified_test_function
def test_NocaseList_imul(testcase, nclist, number, exp_result):
    """
    Test function for NocaseList.__imul__() / ncl *= num
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)
    nclist_copy_id = id(nclist_copy)

    # The code to be tested
    nclist_copy *= number

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the object has been changed in place
    assert id(nclist_copy) == nclist_copy_id

    assert_equal(nclist_copy, exp_result)


TESTCASES_NOCASELIST_REVERSE = [

    # Testcases for NocaseList.reverse()
    # Testcases for NocaseList.__reversed__() / reversed(ncl)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * exp_result: Expected result of the test function, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list",
        dict(
            nclist=NocaseList(),
            exp_result=NocaseList(),
        ),
        None, None, True
    ),
    (
        "List with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            exp_result=NocaseList(['Cat', 'Dog']),
        ),
        None, None, True
    ),
    (
        "List with three items",
        dict(
            nclist=NocaseList(['Dog', 'Cat', 'Kitten']),
            exp_result=NocaseList(['Kitten', 'Cat', 'Dog']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_REVERSE)
@simplified_test_function
def test_NocaseList_reversed(testcase, nclist, exp_result):
    """
    Test function for NocaseList.__reversed__() / reversed(ncl)
    """

    org_nclist = NocaseList(nclist)

    # The code to be tested
    result = reversed(nclist)

    # Verify the input NocaseList object has not changed
    assert_equal(nclist, org_nclist)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_equal(result, exp_result)


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_REVERSE)
@simplified_test_function
def test_NocaseList_reverse(testcase, nclist, exp_result):
    """
    Test function for NocaseList.reverse()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    # pylint: disable=assignment-from-no-return
    result = nclist_copy.reverse()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result is None
    assert_equal(nclist_copy, exp_result)


TESTCASES_NOCASELIST_COMPARE = [

    # Testcases for NocaseList.__eq__(), __ne__()
    # Testcases for NocaseList.__gt__(), __le__()
    # Testcases for NocaseList.__lt__(), __ge__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: NocaseList object #1 to use.
    #   * obj2: NocaseList object #2 to use.
    #   * exp_eq: Expected result of __eq__() and !__ne__()
    #   * exp_gt: Expected result of __gt__() and !__le__()
    #   * exp_lt: Expected result of __lt__() and !__ge__()
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty lists",
        dict(
            obj1=NocaseList(),
            obj2=NocaseList(),
            exp_eq=True,
            exp_gt=False,
            exp_lt=False,
        ),
        None, None, True
    ),
    (
        "Empty lists (list)",
        dict(
            obj1=NocaseList(),
            obj2=[],
            exp_eq=True,
            exp_gt=False,
            exp_lt=False,
        ),
        None, None, True
    ),
    (
        "Lists with one item, that are excase-sensitively equal",
        dict(
            obj1=NocaseList(['Cat']),
            obj2=NocaseList(['Cat']),
            exp_eq=True,
            exp_gt=False,
            exp_lt=False,
        ),
        None, None, True
    ),
    (
        "Lists with one item, that are excase-sensitively equal (list)",
        dict(
            obj1=NocaseList(['Cat']),
            obj2=list(['Cat']),
            exp_eq=True,
            exp_gt=False,
            exp_lt=False,
        ),
        None, None, True
    ),
    (
        "Lists with one item, that are case-insensitively equal",
        dict(
            obj1=NocaseList(['Cat']),
            obj2=NocaseList(['caT']),
            exp_eq=not TEST_AGAINST_LIST,
            exp_gt=False,
            exp_lt=TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "Lists with one item, that are case-insensitively equal (list)",
        dict(
            obj1=NocaseList(['Cat']),
            obj2=list(['caT']),
            exp_eq=not TEST_AGAINST_LIST,
            exp_gt=False,
            exp_lt=TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "Lists with one item, that are case-insensitively lt but "
        "case-sensitively gt",
        dict(
            obj1=NocaseList(['cat']),
            obj2=NocaseList(['Dog']),
            exp_eq=False,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "Lists with one item, that are case-insensitively lt but "
        "case-sensitively gt (list)",
        dict(
            obj1=NocaseList(['cat']),
            obj2=list(['Dog']),
            exp_eq=False,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with one item vs list with two items, first item is "
        "case-insensitively equal",
        dict(
            obj1=NocaseList(['cat']),
            obj2=NocaseList(['Cat', 'Dog']),
            exp_eq=False,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with one item vs list with two items, first item is "
        "case-insensitively equal (list)",
        dict(
            obj1=NocaseList(['cat']),
            obj2=list(['Cat', 'Dog']),
            exp_eq=False,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with one item vs list with two items, first item is not equal",
        dict(
            obj1=NocaseList(['cat']),
            obj2=NocaseList(['Kitten', 'Dog']),
            exp_eq=False,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with one item vs list with two items, first item is not equal "
        "(list)",
        dict(
            obj1=NocaseList(['cat']),
            obj2=list(['Kitten', 'Dog']),
            exp_eq=False,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=not TEST_AGAINST_LIST,
        ),
        None, None, True
    ),
    (
        "List with two items, both case-insensitively equal and in same order",
        dict(
            obj1=NocaseList(['Cat', 'Dog']),
            obj2=NocaseList(['CAT', 'DOG']),
            exp_eq=not TEST_AGAINST_LIST,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=False,
        ),
        None, None, True
    ),
    (
        "List with two items, both case-insensitively equal and in same order "
        "(list)",
        dict(
            obj1=NocaseList(['Cat', 'Dog']),
            obj2=list(['CAT', 'DOG']),
            exp_eq=not TEST_AGAINST_LIST,
            exp_gt=TEST_AGAINST_LIST,
            exp_lt=False,
        ),
        None, None, True
    ),
    (
        "List with two items, both case-insensitively equal but in different "
        "order",
        dict(
            obj1=NocaseList(['Cat', 'Dog']),
            obj2=NocaseList(['DOG', 'CAT']),
            exp_eq=False,
            exp_gt=False,
            exp_lt=True,
        ),
        None, None, True
    ),
    (
        "List with two items, both case-insensitively equal but in different "
        "order (list)",
        dict(
            obj1=NocaseList(['Cat', 'Dog']),
            obj2=list(['DOG', 'CAT']),
            exp_eq=False,
            exp_gt=False,
            exp_lt=True,
        ),
        None, None, True
    ),
]

TESTCASES_NOCASELIST_EQUAL = [

    # Testcases for NocaseList.__eq__(), __ne__()
    (
        "List compared with integer (valid and False)",
        dict(
            obj1=NocaseList(['Cat', 'Dog']),
            obj2=42,
            exp_eq=False,
            exp_gt=None,  # not used
            exp_lt=None,  # not used
        ),
        None, None, True
    ),
    (
        "Integer compared with list (valid and False)",
        dict(
            obj1=42,
            obj2=NocaseList(['Cat', 'Dog']),
            exp_eq=False,
            exp_gt=None,  # not used
            exp_lt=None,  # not used
        ),
        None, None, True
    ),
]

TESTCASES_NOCASELIST_ORDER = [

    # Testcases for NocaseList.__gt__(), __le__()
    # Testcases for NocaseList.__lt__(), __ge__()
    (
        "List compared with integer (invalid)",
        dict(
            obj1=NocaseList(['Cat', 'Dog']),
            obj2=42,
            exp_eq=None,  # not used
            exp_gt=None,
            exp_lt=None,
        ),
        TypeError, None, True
    ),
    (
        "Integer compared with list (invalid)",
        dict(
            obj1=42,
            obj2=NocaseList(['Cat', 'Dog']),
            exp_eq=None,  # not used
            exp_gt=None,
            exp_lt=None,
        ),
        TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COMPARE + TESTCASES_NOCASELIST_EQUAL)  # type: ignore
@simplified_test_function
def test_NocaseList_eq(testcase, obj1, obj2, exp_eq, exp_gt, exp_lt):
    # pylint: disable=unused-argument
    """
    Test function for NocaseList.__eq__()
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    eq1 = (obj1 == obj2)
    eq2 = (obj2 == obj1)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert eq1 == exp_eq
    assert eq2 == exp_eq


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COMPARE + TESTCASES_NOCASELIST_EQUAL)  # type: ignore
@simplified_test_function
def test_NocaseList_ne(testcase, obj1, obj2, exp_eq, exp_gt, exp_lt):
    # pylint: disable=unused-argument
    """
    Test function for NocaseList.__ne__()
    """

    exp_ne = not exp_eq

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    ne1 = (obj1 != obj2)
    ne2 = (obj2 != obj1)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert ne1 == exp_ne
    assert ne2 == exp_ne


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COMPARE + TESTCASES_NOCASELIST_ORDER)  # type: ignore
@simplified_test_function
def test_NocaseList_gt(testcase, obj1, obj2, exp_eq, exp_gt, exp_lt):
    # pylint: disable=unused-argument
    """
    Test function for NocaseList.__gt__()
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    gt = (obj1 > obj2)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert gt == exp_gt


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COMPARE + TESTCASES_NOCASELIST_ORDER)  # type: ignore
@simplified_test_function
def test_NocaseList_lt(testcase, obj1, obj2, exp_eq, exp_gt, exp_lt):
    # pylint: disable=unused-argument
    """
    Test function for NocaseList.__lt__()
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    lt = (obj1 < obj2)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert lt == exp_lt


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COMPARE + TESTCASES_NOCASELIST_ORDER)  # type: ignore
@simplified_test_function
def test_NocaseList_ge(testcase, obj1, obj2, exp_eq, exp_gt, exp_lt):
    # pylint: disable=unused-argument
    """
    Test function for NocaseList.__ge__()
    """

    exp_ge = not exp_lt

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    ge = (obj1 >= obj2)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert ge == exp_ge


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COMPARE + TESTCASES_NOCASELIST_ORDER)  # type: ignore
@simplified_test_function
def test_NocaseList_le(testcase, obj1, obj2, exp_eq, exp_gt, exp_lt):
    # pylint: disable=unused-argument
    """
    Test function for NocaseList.__le__()
    """

    exp_le = not exp_gt

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    le = (obj1 <= obj2)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert le == exp_le


TESTCASES_NOCASELIST_COUNT = [

    # Testcases for NocaseList.count()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * value: Value parameter for the test function.
    #   * exp_result: Expected result of the test function.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with integer value (no casefold)",
        dict(
            nclist=NocaseList(),
            value=1234,
            exp_result=0 if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Empty list, with non-empty string value",
        dict(
            nclist=NocaseList(),
            value='Cat',
            exp_result=0,
        ),
        None, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with integer value (no casefold)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value=1234,
            exp_result=0 if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "List with two items, with non-matching empty string value",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='',
            exp_result=0,
        ),
        None, None, True
    ),
    (
        "List with two items, with non-matching string value",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='Kitten',
            exp_result=0,
        ),
        None, None, True
    ),
    (
        "List with two items, with value matching one list item "
        "(case-insensitively)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='caT',
            exp_result=0 if TEST_AGAINST_LIST else 1,
        ),
        None, None, True
    ),
    (
        "List with four items, with value matching three list items "
        "(case-insensitively)",
        dict(
            nclist=NocaseList(['Cat', 'Dog', 'Cat', 'cat']),
            value='caT',
            exp_result=0 if TEST_AGAINST_LIST else 3,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COUNT)
@simplified_test_function
def test_NocaseList_count(testcase, nclist, value, exp_result):
    """
    Test function for NocaseList.count()
    """

    # The code to be tested
    result = nclist.count(value)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result


TESTCASES_NOCASELIST_COPY = [

    # Testcases for NocaseList.copy()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list",
        dict(
            nclist=NocaseList(),
        ),
        None, None, True
    ),
    (
        "List with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_COPY)
@simplified_test_function
def test_NocaseList_copy(testcase, nclist):
    """
    Test function for NocaseList.copy()
    """

    # The code to be tested
    nclist_copy = nclist.copy()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the result is a different object
    assert id(nclist_copy) != id(nclist)

    assert_equal(nclist_copy, nclist)  # Uses NocaseList equality


TESTCASES_NOCASELIST_CLEAR = [

    # Testcases for NocaseList.clear()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list",
        dict(
            nclist=NocaseList(),
        ),
        None, None, True
    ),
    (
        "List with two items",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_CLEAR)
@simplified_test_function
def test_NocaseList_clear(testcase, nclist):
    """
    Test function for NocaseList.clear()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    # pylint: disable=assignment-from-no-return
    result = nclist_copy.clear()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseList features, but that is
    # unavoidable if we want to work through the public interface:

    assert result is None

    # The following line uses NocaseList len
    assert len(nclist_copy) == 0  # pylint: disable=len-as-condition


TESTCASES_NOCASELIST_INDEX = [

    # Testcases for NocaseList.index()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * args: Positional arguments for the test function.
    #   * exp_result: Expected result of the test function.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with integer value (no casefold)",
        dict(
            nclist=NocaseList(),
            args=(1234,),
            exp_result=None,
        ),
        ValueError if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Empty list, with non-existing string value (not in list)",
        dict(
            nclist=NocaseList(),
            args=('Cat',),
            exp_result=None
        ),
        ValueError, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with integer value (no casefold)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(1234,),
            exp_result=None,
        ),
        ValueError if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "List with two items, with non-existing empty string value",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('',),
            exp_result=NocaseList(['Dog', 'Cat', '']),
        ),
        ValueError, None, True
    ),
    (
        "List with two items, with case-insensitively existing string value "
        "at index 0",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('doG',),
            exp_result=None if TEST_AGAINST_LIST else 0,
        ),
        ValueError if TEST_AGAINST_LIST else None, None, True
    ),
    (
        "List with two items, with case-insensitively existing string value "
        "at index 1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('cAt',),
            exp_result=None if TEST_AGAINST_LIST else 1,
        ),
        ValueError if TEST_AGAINST_LIST else None, None, True
    ),
    (
        "List with two items, with start at 1 and string value of index 0",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('doG', 1),
            exp_result=None,
        ),
        ValueError, None, True
    ),
    (
        "List with two items, with start at 1 and string value of index 1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('caT', 1),
            exp_result=None if TEST_AGAINST_LIST else 1,
        ),
        ValueError if TEST_AGAINST_LIST else None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_INDEX)
@simplified_test_function
def test_NocaseList_index(testcase, nclist, args, exp_result):
    """
    Test function for NocaseList.index()
    """

    # The code to be tested
    result = nclist.index(*args)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result


TESTCASES_NOCASELIST_APPEND = [

    # Testcases for NocaseList.append()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * value: Value parameter for the test function.
    #   * exp_nclist: Expected NocaseList after being updated.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with integer value (no casefold)",
        dict(
            nclist=NocaseList(),
            value=1234,
            exp_nclist=NocaseList([1234]) if TEST_AGAINST_LIST \
            else NocaseList(),
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Empty list, with non-empty string value",
        dict(
            nclist=NocaseList(),
            value='Cat',
            exp_nclist=NocaseList(['Cat']),
        ),
        None, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with integer value (no casefold)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value=1234,
            exp_nclist=NocaseList(['Dog', 'Cat', 1234]) if TEST_AGAINST_LIST \
            else NocaseList(['Dog', 'Cat']),
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "List with two items, with empty string value",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='',
            exp_nclist=NocaseList(['Dog', 'Cat', '']),
        ),
        None, None, True
    ),
    (
        "List with two items, with non-empty string value",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            value='Newbie',
            exp_nclist=NocaseList(['Dog', 'Cat', 'Newbie']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_APPEND)
@simplified_test_function
def test_NocaseList_append(testcase, nclist, value, exp_nclist):
    """
    Test function for NocaseList.append()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    # pylint: disable=assignment-from-no-return
    result = nclist_copy.append(value)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result is None
    assert_equal(nclist_copy, exp_nclist)


TESTCASES_NOCASELIST_EXTEND = [

    # Testcases for NocaseList.extend()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * values: List of values for the test function.
    #   * exp_nclist: Expected NocaseList after being updated.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, with empty tuple",
        dict(
            nclist=NocaseList(),
            values=(),
            exp_nclist=NocaseList(),
        ),
        None, None, True
    ),
    (
        "Empty list, with integer value (no casefold)",
        dict(
            nclist=NocaseList(),
            values=[1234],
            exp_nclist=[1234] if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Empty list, with list containing empty string and non-empty string",
        dict(
            nclist=NocaseList(),
            values=['', 'Cat'],
            exp_nclist=NocaseList(['', 'Cat']),
        ),
        None, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, with empty tuple",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            values=(),
            exp_nclist=NocaseList(['Dog', 'Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, with integer value (no casefold)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            values=[1234],
            exp_nclist=['Dog', 'Cat', 1234] if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "List with two items, with empty string value",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            values=[''],
            exp_nclist=NocaseList(['Dog', 'Cat', '']),
        ),
        None, None, True
    ),
    (
        "List with two items, with non-empty string value",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            values=['Newbie'],
            exp_nclist=NocaseList(['Dog', 'Cat', 'Newbie']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_EXTEND)
@simplified_test_function
def test_NocaseList_extend(testcase, nclist, values, exp_nclist):
    """
    Test function for NocaseList.extend()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    # pylint: disable=assignment-from-no-return
    result = nclist_copy.extend(values)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result is None
    assert_equal(nclist_copy, exp_nclist)


TESTCASES_NOCASELIST_INSERT = [

    # Testcases for NocaseList.insert()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * args: Positional arguments for the test function.
    #   * exp_nclist: Expected NocaseList after being updated.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, inserting integer value before index 0 (no casefold)",
        dict(
            nclist=NocaseList(),
            args=(0, 1234),
            exp_nclist=[1234] if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "Empty list, inserting string value before index 0",
        dict(
            nclist=NocaseList(),
            args=(0, 'Newbie'),
            exp_nclist=NocaseList(['Newbie']),
        ),
        None, None, True
    ),
    (
        "Empty list, inserting string value before index 1",
        dict(
            nclist=NocaseList(),
            args=(1, 'Newbie'),
            exp_nclist=NocaseList(['Newbie']),
        ),
        None, None, True
    ),
    (
        "Empty list, inserting string value before index -1",
        dict(
            nclist=NocaseList(),
            args=(-1, 'Newbie'),
            exp_nclist=NocaseList(['Newbie']),
        ),
        None, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, inserting integer value before index 0 "
        "(no casefold)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(0, 1234),
            exp_nclist=[1234, 'Dog', 'Cat'] if TEST_AGAINST_LIST else None,
        ),
        None if TEST_AGAINST_LIST else AttributeError, None, True
    ),
    (
        "List with two items, inserting string value before index 0",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(0, 'Newbie'),
            exp_nclist=NocaseList(['Newbie', 'Dog', 'Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, inserting string value before index 1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(1, 'Newbie'),
            exp_nclist=NocaseList(['Dog', 'Newbie', 'Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, inserting string value before index 2",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(2, 'Newbie'),
            exp_nclist=NocaseList(['Dog', 'Cat', 'Newbie']),
        ),
        None, None, True
    ),
    (
        "List with two items, inserting string value before index 3",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(3, 'Newbie'),
            exp_nclist=NocaseList(['Dog', 'Cat', 'Newbie']),
        ),
        None, None, True
    ),
    (
        "List with two items, inserting string value before index -1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(-1, 'Newbie'),
            exp_nclist=NocaseList(['Dog', 'Newbie', 'Cat']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_INSERT)
@simplified_test_function
def test_NocaseList_insert(testcase, nclist, args, exp_nclist):
    """
    Test function for NocaseList.insert()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    # pylint: disable=assignment-from-no-return
    result = nclist_copy.insert(*args)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result is None
    assert_equal(nclist_copy, exp_nclist)


TESTCASES_NOCASELIST_POP = [

    # Testcases for NocaseList.pop()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * args: Positional arguments for the test function.
    #   * exp_result: Expected result from the test function.
    #   * exp_nclist: Expected NocaseList after being updated.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, pop index 0 (does not exist)",
        dict(
            nclist=NocaseList(),
            args=(0,),
            exp_result=None,
            exp_nclist=None,
        ),
        IndexError, None, True
    ),
    (
        "Empty list, pop default index -1 (does not exist)",
        dict(
            nclist=NocaseList(),
            args=(),
            exp_result=None,
            exp_nclist=None,
        ),
        IndexError, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, pop index 0",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(0,),
            exp_result='Dog',
            exp_nclist=NocaseList(['Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, pop index 1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(1,),
            exp_result='Cat',
            exp_nclist=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "List with two items, pop index -1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(-1,),
            exp_result='Cat',
            exp_nclist=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "List with two items, pop default index -1",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=(),
            exp_result='Cat',
            exp_nclist=NocaseList(['Dog']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_POP)
@simplified_test_function
def test_NocaseList_pop(testcase, nclist, args, exp_result, exp_nclist):
    """
    Test function for NocaseList.pop()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    result = nclist_copy.pop(*args)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result
    assert_equal(nclist_copy, exp_nclist)


TESTCASES_NOCASELIST_REMOVE = [

    # Testcases for NocaseList.remove()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * args: Positional arguments for the test function.
    #   * exp_nclist: Expected NocaseList after being updated.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, remove some value (does not exist)",
        dict(
            nclist=NocaseList(),
            args=('Cat',),
            exp_nclist=None,
        ),
        ValueError, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, remove first",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('Dog',),
            exp_nclist=NocaseList(['Cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, remove second",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('Cat',),
            exp_nclist=NocaseList(['Dog']),
        ),
        None, None, True
    ),
    (
        "List with two items, remove non-existing (not found)",
        dict(
            nclist=NocaseList(['Dog', 'Cat']),
            args=('Kitten',),
            exp_nclist=None,
        ),
        ValueError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_REMOVE)
@simplified_test_function
def test_NocaseList_remove(testcase, nclist, args, exp_nclist):
    """
    Test function for NocaseList.remove()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    # pylint: disable=assignment-from-no-return
    result = nclist_copy.remove(*args)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result is None
    assert_equal(nclist_copy, exp_nclist)


TESTCASES_NOCASELIST_SORT = [

    # Testcases for NocaseList.sort()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    #   * kwargs: Keyword arguments for the test function.
    #   * exp_nclist: Expected NocaseList after being updated.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseList
    (
        "Empty list, default sort",
        dict(
            nclist=NocaseList(),
            kwargs={},
            exp_nclist=NocaseList(),
        ),
        None, None, True
    ),

    # Non-empty NocaseList
    (
        "List with two items, case-insensitively unsorted, default sort key",
        dict(
            nclist=NocaseList(['Dog', 'cat']),
            kwargs={},
            exp_nclist=['Dog', 'cat'] if TEST_AGAINST_LIST \
            else NocaseList(['cat', 'Dog']),
        ),
        None, None, True
    ),
    (
        "List with two items, case-insensitively unsorted, default sort key, "
        "descending",
        dict(
            nclist=NocaseList(['cat', 'Dog']),
            kwargs=dict(
                reverse=True,
            ),
            exp_nclist=['cat', 'Dog'] if TEST_AGAINST_LIST \
            else NocaseList(['Dog', 'cat']),
        ),
        None, None, True
    ),
    (
        "List with two items, case-insensitively unsorted, with sort key",
        dict(
            nclist=NocaseList(['BUdgie', 'cat']),
            kwargs=dict(
                key=lambda x: x[1],  # sort by second character
            ),
            exp_nclist=['BUdgie', 'cat'] if TEST_AGAINST_LIST \
            else NocaseList(['cat', 'BUdgie']),
        ),
        None, None, True
    ),
    (
        "List with two items, case-insensitively unsorted, with sort key, "
        "descending",
        dict(
            nclist=NocaseList(['cat', 'BUdgie']),
            kwargs=dict(
                key=lambda x: x[1],  # sort by second character
                reverse=True,
            ),
            exp_nclist=['cat', 'BUdgie'] if TEST_AGAINST_LIST \
            else NocaseList(['BUdgie', 'cat']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_SORT)
@simplified_test_function
def test_NocaseList_sort(testcase, nclist, kwargs, exp_nclist):
    """
    Test function for NocaseList.sort()
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # The code to be tested
    # pylint: disable=assignment-from-no-return
    result = nclist_copy.sort(**kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result is None
    assert_equal(nclist_copy, exp_nclist)


TESTCASES_NOCASELIST_PICKLE = [

    # Testcases for pickling and unpickling NocaseList objects

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * nclist: NocaseList object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty list",
        dict(
            nclist=NocaseList(),
        ),
        None, None, True
    ),
    (
        "List with two items",
        dict(
            nclist=NocaseList(['Dog', 'cat']),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASELIST_PICKLE)
@simplified_test_function
def test_NocaseList_pickle(testcase, nclist):
    """
    Test function for pickling and unpickling NocaseList objects
    """

    # Don't change the testcase data, but a copy
    nclist_copy = NocaseList(nclist)

    # Pickle the object
    pkl = pickle.dumps(nclist_copy)

    del nclist_copy

    # Unpickle the object
    nclist2 = pickle.loads(pkl)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert_equal(nclist2, nclist)


def test_casefold_override():
    """
    Test function for overriding the casefold method.
    """

    if TEST_AGAINST_LIST:
        pytest.skip("The override test does not support testing with list")

    class MyNocaseList(NocaseList):
        "Test class that overrides the casefold method"

        @staticmethod
        def __casefold__(value):
            return unicodedata.normalize('NFKD', value).casefold()

    nclist = MyNocaseList()

    # Add item with combined Unicode character
    nclist.append("\u00C7")

    # Look up item with combination sequence
    assert "c\u0327" in nclist



#!/usr/bin/env python
"""
Python setup script for the nocaselist project.
"""

import sys
import os
import io
import re
from typing import Optional

import setuptools  # type: ignore


def get_version(version_file):
    """
    Execute the specified version file and return the value of the __version__
    global variable that is set in the version file.

    Note: Make sure the version file does not depend on any packages in the
    requirements list of this package (otherwise it cannot be executed in
    a fresh Python environment).
    """
    with io.open(version_file, 'r', encoding='utf-8') as fp:
        version_source = fp.read()
    _globals = {}
    exec(version_source, _globals)  # pylint: disable=exec-used
    return _globals['__version__']


def get_requirements(requirements_file):
    """
    Parse the specified requirements file and return a list of its non-empty,
    non-comment lines. The returned lines are without any trailing newline
    characters.
    """
    with io.open(requirements_file, 'r', encoding='utf-8') as fp:
        lines = fp.readlines()
    reqs = []
    for line in lines:
        line = line.strip('\n')
        if not line.startswith('#') and line != '':
            reqs.append(line)
    return reqs


def read_file(a_file):
    """
    Read the specified file and return its content as one string.
    """
    with io.open(a_file, 'r', encoding='utf-8') as fp:
        content = fp.read()
    return content


class PytestCommand(setuptools.Command):
    """
    Base class for setup.py commands for executing tests for this package
    using pytest.

    Note on the class name: Because distutils.dist._show_help() shows the class
    name for the setup.py command name instead of invoking get_command_name(),
    the classes that get registered as commands must have the command name.
    """

    description: Optional[str] = None  # Set by subclass
    my_test_dirs: Optional[list] = None  # Set by subclass

    user_options = [
        (
            'pytest-options=',  # '=' indicates it requires an argument
            None,  # no short option
            "additional options for pytest, as one argument"
        ),
    ]

    def initialize_options(self):
        """
        Standard method called by setup to initialize options for the command.
        """
        # pylint: disable=attribute-defined-outside-init
        self.test_opts = None
        self.test_dirs = None
        self.pytest_options = None
        # pylint: enable=attribute-defined-outside-init

    def finalize_options(self):
        """
        Standard method called by setup to finalize options for the command.
        """
        # pylint: disable=attribute-defined-outside-init
        self.test_opts = [
            '--color=yes',
            '-s',
            '-W', 'default',
            '-W', 'ignore::PendingDeprecationWarning',
        ]
        if sys.version_info[0] == 3:
            self.test_opts.extend([
                '-W', 'ignore::ResourceWarning',
            ])
        self.test_dirs = self.my_test_dirs
        # pylint: enable=attribute-defined-outside-init

    def run(self):
        """
        Standard method called by setup to execute the command.
        """

        # deferred import so install does not depend on it
        import pytest  # pylint: disable=import-outside-toplevel

        args = self.test_opts
        if self.pytest_options:
            args.extend(self.pytest_options.split(' '))
        args.extend(self.test_dirs)

        if self.dry_run:
            self.announce("Dry-run: pytest {}".format(' '.join(args)), level=1)
            return 0

        self.announce("pytest {}".format(' '.join(args)), level=1)
        rc = pytest.main(args)
        return rc


class test(PytestCommand):
    # pylint: disable=invalid-name
    """
    Setup.py command for executing unit and function tests.
    """
    description = "nocaselist: Run unit tests using pytest"
    my_test_dirs = ['tests/unittest']


# pylint: disable=invalid-name
requirements = get_requirements('requirements.txt')
install_requires = [req for req in requirements
                    if req and not re.match(r'[^:]+://', req)]
dependency_links = [req for req in requirements
                    if req and re.match(r'[^:]+://', req)]
test_requirements = get_requirements('test-requirements.txt')

package_version = get_version(os.path.join('nocaselist', '_version.py'))

# Docs on setup():
# * https://docs.python.org/3/distutils/apiref.html?
#   highlight=setup#distutils.core.setup
# * https://setuptools.readthedocs.io/en/latest/setuptools.html#
#   new-and-changed-setup-keywords
setuptools.setup(
    name='nocaselist',
    version=package_version,
    packages=[
        'nocaselist',
    ],
    package_data={
        'nocaselist': ['py.typed']
    },
    include_package_data=True,  # Includes MANIFEST.in files into sdist (only)
    scripts=[
        # add any scripts
    ],
    install_requires=install_requires,
    dependency_links=dependency_links,
    extras_require={
        "test": test_requirements,
    },
    cmdclass={
        'test': test,
    },
    description="A case-insensitive list for Python",
    long_description=read_file('README.rst'),
    long_description_content_type='text/x-rst',
    license="Apache Software License 2.0",
    author="Andreas Maier",
    author_email='andreas.r.maier@gmx.de',
    maintainer="Andreas Maier",
    maintainer_email='andreas.r.maier@gmx.de',
    url='https://github.com/pywbem/nocaselist',
    project_urls={
        'Bug Tracker': 'https://github.com/pywbem/nocaselist/issues',
        'Documentation': 'https://nocaselist.readthedocs.io/en/latest/',
        'Source Code': 'https://github.com/pywbem/nocaselist',
    },

    options={'bdist_wheel': {'universal': True}},
    zip_safe=True,  # This package can safely be installed from a zip file
    platforms='any',

    # Keep these Python versions in sync with nocaselist/__init__.py
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



"""
Version of the nocaselist package.
"""

#: The full version of this package including any development levels, as a
#: string.
#:
#: Possible formats for this version string are:
#:
#: * "M.N.P.dev1": Development level 1 of a not yet released version M.N.P
#: * "M.N.P": A released version M.N.P
__version__: str = '2.0.0'



"""
Test import and versioning of the package.
"""


def test_import():
    """
    Test import of the package.
    """
    import nocaselist  # noqa: F401 pylint: disable=import-outside-toplevel
    assert nocaselist


def test_versioning():
    """
    Test import of the package.
    """
    import nocaselist  # noqa: F401 pylint: disable=import-outside-toplevel
    assert nocaselist.__version__
